/******************************************************* 
*    Copyright(c) 2020 Braveridge Co., Ltd. All Rights Reserved                                       
*******************************************************/
/******************************************************* 
*    include                                     
*******************************************************/
#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_sdm.h"
#include "app_error.h"
#include "ble.h"
#include "ble_err.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "app_timer.h"
#include "fds.h"
#include "nrf_ble_gatt.h"
#include "ble_conn_state.h"
#include "nrf_pwr_mgmt.h"

//gpio
#include "nrf_gpio.h"

//log
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

//mic
#include "nrf_drv_pdm.h"
//speaker
#include "nrf_drv_i2s.h"

/******************************************************* 
*    define                                     
*******************************************************/
#define DEVICE_NAME                         "MicSpeakerSample"                            /**< Name of device. Will be included in the advertising data. */
#define MANUFACTURER_NAME                   "Braveridge"                   /**< Manufacturer. Will be passed to Device Information Service. */
#define APP_ADV_INTERVAL                    300                                     /**< The advertising interval (in units of 0.625 ms. This value corresponds to 187.5 ms). */

#define APP_ADV_DURATION                    18000                                   /**< The advertising duration (180 seconds) in units of 10 milliseconds. */

#define APP_BLE_CONN_CFG_TAG                1                                       /**< A tag identifying the SoftDevice BLE configuration. */
#define APP_BLE_OBSERVER_PRIO               3                                       /**< Application's BLE observer priority. You shouldn't need to modify this value. */

#define MIN_CONN_INTERVAL                   MSEC_TO_UNITS(400, UNIT_1_25_MS)        /**< Minimum acceptable connection interval (0.4 seconds). */
#define MAX_CONN_INTERVAL                   MSEC_TO_UNITS(650, UNIT_1_25_MS)        /**< Maximum acceptable connection interval (0.65 second). */
#define SLAVE_LATENCY                       0                                       /**< Slave latency. */
#define CONN_SUP_TIMEOUT                    MSEC_TO_UNITS(4000, UNIT_10_MS)         /**< Connection supervisory timeout (4 seconds). */

#define FIRST_CONN_PARAMS_UPDATE_DELAY      APP_TIMER_TICKS(5000)                   /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY       APP_TIMER_TICKS(30000)                  /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT        3                                       /**< Number of attempts before giving up the connection parameter negotiation. */

/* speaker port
 *
 * 今回は以下のように配線しました。
 * SD_MODE: P1.07
 * DIN    : P1.05
 * BCLK   : P1.02
 * LRCLK  : P1.03
 *
 */
#define PORT_SPEAKER_ENABLE                 NRF_GPIO_PIN_MAP(1,7)
#define PORT_I2S_SDOUT                      NRF_GPIO_PIN_MAP(1,5)
#define PORT_I2S_SCK                        NRF_GPIO_PIN_MAP(1,2)
#define PORT_I2S_LRCK                       NRF_GPIO_PIN_MAP(1,3)

/* mic port
 *
 * 今回は以下のように配線しました。
 * Mic DATA: P1.01
 * Mic CLK : P1.04
 *
 */
#define	PORT_MIC_CLK                        NRF_GPIO_PIN_MAP(1,4)
#define PORT_MIC_DATA                       NRF_GPIO_PIN_MAP(1,1)

#define I2S_DATA_BLOCK_WORDS                512
#define CONFIG_PDM_BUFFER_SIZE_SAMPLES      4096

#define MIC_DATA_IGNORE_CNT                 5

#define DEAD_BEEF                           0xDEADBEEF                              /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

/******************************************************* 
*    instance                                     
*******************************************************/
NRF_BLE_GATT_DEF(m_gatt);                                           /**< GATT module instance. */
BLE_ADVERTISING_DEF(m_advertising);                                 /**< Advertising module instance. */

/******************************************************* 
*    data                                     
*******************************************************/
static uint16_t m_conn_handle         = BLE_CONN_HANDLE_INVALID;    /**< Handle of the current connection. */
static bool is_speaker_started = false;
static int16_t	m_pdm_buff[CONFIG_PDM_BUFFER_SIZE_SAMPLES];
static uint8_t data_buf[CONFIG_PDM_BUFFER_SIZE_SAMPLES*2] = {0};
static uint32_t m_buffer_rx[2][CONFIG_PDM_BUFFER_SIZE_SAMPLES] = {{0}};
static uint32_t m_buffer_tx[2][CONFIG_PDM_BUFFER_SIZE_SAMPLES] = {{0}};
static uint8_t pdm_handler_count = 0;


/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num   Line number of the failing ASSERT call.
 * @param[in] file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
	app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


/**@brief Function for starting advertising.
 */
void advertising_start(void)
{
	ret_code_t err_code;

	err_code = ble_advertising_start(&m_advertising, BLE_ADV_MODE_FAST);
	APP_ERROR_CHECK(err_code);
}


/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module. This creates and starts application timers.
 */
static void timers_init(void)
{
	ret_code_t err_code;

	// Initialize timer module.
	err_code = app_timer_init();
	APP_ERROR_CHECK(err_code);

	// Create timers.
}


/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
	ret_code_t              err_code;
	ble_gap_conn_params_t   gap_conn_params;
	ble_gap_conn_sec_mode_t sec_mode;

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

	err_code = sd_ble_gap_device_name_set(&sec_mode,
										  (const uint8_t *)DEVICE_NAME,
										  strlen(DEVICE_NAME));
	APP_ERROR_CHECK(err_code);

	err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_HEART_RATE_SENSOR_HEART_RATE_BELT);
	APP_ERROR_CHECK(err_code);

	memset(&gap_conn_params, 0, sizeof(gap_conn_params));

	gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
	gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
	gap_conn_params.slave_latency     = SLAVE_LATENCY;
	gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

	err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
	APP_ERROR_CHECK(err_code);
}


/**@brief GATT module event handler.
 */
static void gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
    if (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED)
    {
        NRF_LOG_INFO("GATT ATT MTU on connection 0x%x changed to %d.",
                     p_evt->conn_handle,
                     p_evt->params.att_mtu_effective);
    }
}


/**@brief Function for initializing the GATT module.
 */
static void gatt_init(void)
{
    ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling advertising events.
 *
 * @details This function will be called for advertising events which are passed to the application.
 *
 * @param[in] ble_adv_evt  Advertising event.
 */
static void on_adv_evt(ble_adv_evt_t ble_adv_evt)
{
	ret_code_t err_code;

	switch (ble_adv_evt)
	{
		case BLE_ADV_EVT_FAST:
			NRF_LOG_INFO("Fast advertising.");
			break;

		case BLE_ADV_EVT_IDLE:
			break;

		default:
			break;
	}
}


/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
	ret_code_t err_code;

	switch (p_ble_evt->header.evt_id)
	{
		case BLE_GAP_EVT_CONNECTED:
			NRF_LOG_INFO("Connected.");
			m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
			break;

		case BLE_GAP_EVT_DISCONNECTED:
			NRF_LOG_INFO("Disconnected, reason %d.",
						  p_ble_evt->evt.gap_evt.params.disconnected.reason);
			m_conn_handle = BLE_CONN_HANDLE_INVALID;
			break;

		case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
		{
			NRF_LOG_DEBUG("PHY update request.");
			ble_gap_phys_t const phys =
			{
				.rx_phys = BLE_GAP_PHY_AUTO,
				.tx_phys = BLE_GAP_PHY_AUTO,
			};
			err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
			APP_ERROR_CHECK(err_code);
		} break;

		case BLE_GATTC_EVT_TIMEOUT:
			// Disconnect on GATT Client timeout event.
			NRF_LOG_DEBUG("GATT Client Timeout.");
			err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
											 BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
			APP_ERROR_CHECK(err_code);
			break;

		case BLE_GATTS_EVT_TIMEOUT:
			// Disconnect on GATT Server timeout event.
			NRF_LOG_DEBUG("GATT Server Timeout.");
			err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
											 BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
			APP_ERROR_CHECK(err_code);
			break;
	
		case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
			NRF_LOG_DEBUG("BLE_GAP_EVT_SEC_PARAMS_REQUEST");
			break;
		
		case BLE_GAP_EVT_AUTH_KEY_REQUEST:
			NRF_LOG_INFO("BLE_GAP_EVT_AUTH_KEY_REQUEST");
			break;

		case BLE_GAP_EVT_LESC_DHKEY_REQUEST:
			NRF_LOG_INFO("BLE_GAP_EVT_LESC_DHKEY_REQUEST");
			break;

		 case BLE_GAP_EVT_AUTH_STATUS:
			 NRF_LOG_INFO("BLE_GAP_EVT_AUTH_STATUS: status=0x%x bond=0x%x lv4: %d kdist_own:0x%x kdist_peer:0x%x",
						  p_ble_evt->evt.gap_evt.params.auth_status.auth_status,
						  p_ble_evt->evt.gap_evt.params.auth_status.bonded,
						  p_ble_evt->evt.gap_evt.params.auth_status.sm1_levels.lv4,
						  *((uint8_t *)&p_ble_evt->evt.gap_evt.params.auth_status.kdist_own),
						  *((uint8_t *)&p_ble_evt->evt.gap_evt.params.auth_status.kdist_peer));
			break;

		default:
			// No implementation needed.
			break;
	}
}


/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
	ret_code_t err_code;

	err_code = nrf_sdh_enable_request();
	APP_ERROR_CHECK(err_code);

	// Configure the BLE stack using the default settings.
	// Fetch the start address of the application RAM.
	uint32_t ram_start = 0;
	err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
	APP_ERROR_CHECK(err_code);

	// Enable BLE stack.
	err_code = nrf_sdh_ble_enable(&ram_start);
	APP_ERROR_CHECK(err_code);

	// Register a handler for BLE events.
	NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}


/**@brief Function for initializing the Advertising functionality.
 */
static void advertising_init(void)
{
	ret_code_t             err_code;
	ble_advertising_init_t init;

	memset(&init, 0, sizeof(init));

	init.advdata.name_type               = BLE_ADVDATA_FULL_NAME;
	init.advdata.flags                   = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;

	init.config.ble_adv_fast_enabled  = true;
	init.config.ble_adv_fast_interval = APP_ADV_INTERVAL;
	init.config.ble_adv_fast_timeout  = APP_ADV_DURATION;

	init.evt_handler = on_adv_evt;

	err_code = ble_advertising_init(&m_advertising, &init);
	APP_ERROR_CHECK(err_code);

	ble_advertising_conn_cfg_tag_set(&m_advertising, APP_BLE_CONN_CFG_TAG);
}


/**@brief Function for initializing the nrf log module.
 */
static void log_init(void)
{
	ret_code_t err_code = NRF_LOG_INIT(NULL);
	APP_ERROR_CHECK(err_code);

	NRF_LOG_DEFAULT_BACKENDS_INIT();
}


/**@brief Function for initializing power management.
 */
static void power_management_init(void)
{
	ret_code_t err_code;
	err_code = nrf_pwr_mgmt_init();
	APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling the idle state (main loop).
 *
 * @details If there is no pending log operation, then sleep until next the next event occurs.
 */
static void idle_state_handle(void)
{
	ret_code_t err_code;

	if (NRF_LOG_PROCESS() == false)
	{
		nrf_pwr_mgmt_run();
	}
}


/******************************************************************
 * Speaker
 *
 * nrfx_i2s.c をDriver内に入れ、sdk_config.h 内の「I2S_ENABLED」を0から1にすることで「nrf_drv_i2s_init」等を使用できるようになります。
 * 
 ******************************************************************/
static void i2s_data_handler(nrf_drv_i2s_buffers_t const * p_released, uint32_t status)
{
	ASSERT(p_released);

        /*drv_pdm_event_handler で取得したデータ(data_buf)をセットする。*/
	memcpy(m_buffer_tx[1], data_buf, CONFIG_PDM_BUFFER_SIZE_SAMPLES);

	nrf_drv_i2s_buffers_t const next_buffers = {
		.p_rx_buffer = m_buffer_rx[1],
		.p_tx_buffer = m_buffer_tx[1],
	};
	APP_ERROR_CHECK(nrf_drv_i2s_next_buffers_set(&next_buffers));
}

static void speaker_init(void)
{
	ret_code_t err_code;

	nrf_gpio_cfg_output(PORT_SPEAKER_ENABLE);
	nrf_gpio_pin_clear(PORT_SPEAKER_ENABLE);

	nrf_drv_i2s_config_t config = NRF_DRV_I2S_DEFAULT_CONFIG;
	config.sdin_pin             = NRF_DRV_I2S_PIN_NOT_USED;
	config.sdout_pin            = PORT_I2S_SDOUT;
	config.sck_pin              = PORT_I2S_SCK;
	config.lrck_pin             = PORT_I2S_LRCK;
	config.mck_pin              = NRF_DRV_I2S_PIN_NOT_USED;
	config.mck_setup            = NRF_I2S_MCK_32MDIV63;
	config.ratio                = NRF_I2S_RATIO_32X;
	config.sample_width         = NRF_I2S_SWIDTH_16BIT;
	config.channels             = NRF_I2S_CHANNELS_LEFT;

	err_code = nrf_drv_i2s_init(&config, i2s_data_handler);
	APP_ERROR_CHECK(err_code);
}

static void speaker_start(void)
{
	ret_code_t err_code;

	if (is_speaker_started == true)
	{
		return;      
	}

	nrf_gpio_pin_set(PORT_SPEAKER_ENABLE);
	is_speaker_started = true;

	nrf_drv_i2s_buffers_t const initial_buffers = {
		.p_tx_buffer = m_buffer_tx[0],
		.p_rx_buffer = m_buffer_rx[0],
	};

	err_code = nrf_drv_i2s_start(&initial_buffers, I2S_DATA_BLOCK_WORDS, 0);
	APP_ERROR_CHECK(err_code);	
}
/******************************************************************
 * Mic
 *
 * nrfx_pwm.c をDriver内に入れ、sdk_config.h 内のPDM_ENABLEDを0から1にすることで「nrf_drv_pdm_init」等を使用できるようになります。
 *
 ******************************************************************/
static void drv_pdm_event_handler(nrf_drv_pdm_evt_t const * const evt)
{
	ret_code_t err_code;

	if (evt->buffer_requested)
	{
		err_code = nrf_drv_pdm_buffer_set(m_pdm_buff, CONFIG_PDM_BUFFER_SIZE_SAMPLES);
		APP_ERROR_CHECK(err_code);

		// 最初はデータにノイズが乗っているから無視する。
		if(pdm_handler_count < MIC_DATA_IGNORE_CNT)
		{
			pdm_handler_count++;
		}
		else
		{
			/*取得したデータをバッファに格納*/
			memcpy(data_buf, evt->buffer_released, CONFIG_PDM_BUFFER_SIZE_SAMPLES*2);

			/*マイクで音を拾ったらスピーカを鳴らす*/
			speaker_start();
		}
	}
}

static void mic_init(void)
{
	ret_code_t err_code;

	nrf_drv_pdm_config_t pdm_cfg = NRF_DRV_PDM_DEFAULT_CONFIG(PORT_MIC_CLK, PORT_MIC_DATA);

	err_code = nrf_drv_pdm_init(&pdm_cfg, drv_pdm_event_handler);
	APP_ERROR_CHECK(err_code);
}

static void mic_start(void)
{
	ret_code_t err_code;

	err_code = nrf_drv_pdm_start();
	APP_ERROR_CHECK(err_code);
}


/**@brief Function for application main entry.
 */
int main(void)
{
	// Initialize.
	log_init();
	timers_init();
	power_management_init();
	ble_stack_init();
	gap_params_init();
	gatt_init();
	advertising_init();

	mic_init();
	speaker_init();

	// Start execution.
	advertising_start();
	mic_start();

	//sdk_config.h 内のNRF_LOG_BACKEND_RTT_ENABLEDを0から1にすることで、ログが表示されるようになります。
	NRF_LOG_INFO("Mic Speaker example started.");

	// Enter main loop.
	for (;;)
	{
		idle_state_handle();
	}
}

